package tree

import (
	"bytes"
	"math/big"

	"github.com/ethereum/go-ethereum/common"
	"github.com/ethereum/go-ethereum/crypto"
)

type MerkleSumTree struct {
	Levels   [][]*MerkleSumNode
	RootNode *MerkleSumNode
}

type MerkleSumNode struct {
	Left  *MerkleSumNode
	Right *MerkleSumNode
	Value Value
}

type Value struct {
	BigValue *big.Int
	Hash     common.Hash
	Raw      []byte
}

func NewMerkleSumTree(datas [][]byte, vals []*big.Int) *MerkleSumTree {
	var nodes []*MerkleSumNode
	var levels [][]*MerkleSumNode

	if len(datas) == 1 {
		return &MerkleSumTree{
			Levels:   levels,
			RootNode: buildToNode(datas[0], vals[0]),
		}
	}

	for i := range datas {
		nodes = append(nodes, buildToNode(datas[i], vals[i]))
	}

	if len(nodes)%2 == 1 {
		// raw = 0x09标识，添加到最后一层
		nodes = append(nodes, &MerkleSumNode{
			Value: Value{
				BigValue: big.NewInt(0),
				Raw:      []byte{0x09},
			},
		})
	}

	countOfDataNodes := len(nodes)
	counterOfLevels := 0
	for countOfDataNodes > 1 {
		if countOfDataNodes%2 == 0 {
			countOfDataNodes = countOfDataNodes / 2
			counterOfLevels++
		} else {
			countOfDataNodes = (countOfDataNodes + 1) / 2
			counterOfLevels++
		}
	}

	levels = [][]*MerkleSumNode{nodes}

	for i := 0; i < counterOfLevels; i++ {
		var level []*MerkleSumNode

		lastNodeIndex := len(nodes) - 1
		for j := 0; j <= lastNodeIndex; j += 2 {
			if j == lastNodeIndex && lastNodeIndex%2 == 0 {
				node := newMerkleSumNode(nodes[j], nil)
				level = append(level, node)
			} else {
				node := newMerkleSumNode(nodes[j], nodes[j+1])
				level = append(level, node)
			}
		}
		nodes = level
		levels = append(levels, level)
	}

	tree := MerkleSumTree{levels, nodes[0]}

	return &tree
}

func newMerkleSumNode(left, right *MerkleSumNode) *MerkleSumNode {
	var node MerkleSumNode
	if right == nil || (right != nil && bytes.Equal(right.Value.Raw, []byte{0x09})) {
		// concatLeftNodeData := append(left.Value.BigValue.Bytes(), left.Value.Hash[:]...)
		// concatRightNodeData := append(big.NewInt(0).Bytes(), (common.Hash{}).Bytes()...)
		// prevHashes := append(concatLeftNodeData)
		node.Value.Hash = left.Value.Hash
		node.Value.BigValue = left.Value.BigValue
	} else {
		concatLeftNodeData := append(left.Value.BigValue.Bytes(), left.Value.Hash[:]...)
		concatRightNodeData := append(right.Value.BigValue.Bytes(), right.Value.Hash[:]...)
		prevHashes := append(concatLeftNodeData, concatRightNodeData...)
		node.Value.Hash = crypto.Keccak256Hash(prevHashes)
		node.Value.BigValue = big.NewInt(0).Add(left.Value.BigValue, right.Value.BigValue)
	}

	node.Left = left
	node.Right = right

	return &node
}

func buildToNode(data []byte, val *big.Int) *MerkleSumNode {
	node := MerkleSumNode{
		Value: Value{
			BigValue: val,
			Hash:     crypto.Keccak256Hash(append(val.Bytes(), data...)),
			Raw:      data,
		},
	}
	return &node
}

func (m *MerkleSumTree) GetRoot() *MerkleSumNode {
	return m.RootNode
}

func (m *MerkleSumTree) FindMerkleSumNode(hash common.Hash) *MerkleSumNode {
	if m.RootNode == nil {
		return nil
	}

	queue := []*MerkleSumNode{m.RootNode}

	for len(queue) > 0 {
		size := len(queue)

		for i := 0; i < size; i++ {
			currentNode := queue[0]
			queue = queue[1:]

			if currentNode.Value.Hash.Hex() == hash.Hex() {
				return currentNode
			}

			if currentNode.Left != nil {
				queue = append(queue, currentNode.Left)
			}
			if currentNode.Right != nil {
				queue = append(queue, currentNode.Right)
			}
		}
	}
	return nil
}

func MerkleSumTreeTraversal(root *MerkleSumNode, depth int) (retNodes [][]Value) {
	if root == nil {
		return nil
	}

	queue := []*MerkleSumNode{root}

	currentDepth := 1

	for len(queue) > 0 {
		size := len(queue)

		layerNodes := make([]Value, 0)
		for i := 0; i < size; i++ {
			currentNode := queue[0]
			queue = queue[1:]

			layerNodes = append(layerNodes, currentNode.Value)

			if currentNode.Left != nil {
				queue = append(queue, currentNode.Left)
			}
			if currentNode.Right != nil {
				queue = append(queue, currentNode.Right)
			}
		}
		retNodes = append(retNodes, layerNodes)
		currentDepth++
		if currentDepth > depth {
			break
		}
	}
	return
}
